"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var dependency_1 = require("./dependency");
require("source-map-support/register");
var async = require("async");
var fs = require("fs");
var pathModule = require("path");
var colors = require("colors/safe");
var Leaf = require("leaf-ts");
var sharedCallbacks_1 = require("./sharedCallbacks");
var useColors = colors;
var ServiceContext = /** @class */ (function () {
    function ServiceContext(settings) {
        if (settings === void 0) { settings = {}; }
        var _this = this;
        this.settings = settings;
        this.BuiltInServiceInstances = {};
        this.services = {};
        this.dependencyManager = new dependency_1.DependencyManager;
        this.supportedLoadExtnames = [".ts", ".js"];
        this.maxInitializingTimeout = 1000 * 5;
        this.isSetup = false;
        this.finishSetup = sharedCallbacks_1.SharedCallbacks.create();
        var BuiltInService = require("./builtInService");
        for (var name_1 in BuiltInService) {
            var Con = BuiltInService[name_1];
            var s = new Con(this.settings);
            this.BuiltInServiceInstances[s.name] = s;
        }
        // When dependency not met but we have that dependency in our built-in service
        this.dependencyManager.onAddDymanicDependency = function (name) {
            var s = _this.BuiltInServiceInstances[name];
            if (s) {
                if (s["asBuiltIn"]) {
                    s["asBuiltIn"];
                }
                _this.register(s);
            }
        };
    }
    ServiceContext.prototype.register = function (s) {
        console.log("Register service", useColors.green(s.name));
        this.dependencyManager.add(s.name, s.dependencies || null);
        this.services[s.name] = s;
    };
    ServiceContext.prototype.loadServices = function (path) {
        var _this = this;
        path = pathModule.resolve(path);
        var files = fs.readdirSync(path);
        files = files.filter(function (item) {
            var ext = pathModule.extname(item);
            if (_this.supportedLoadExtnames.indexOf(ext) < 0) {
                return false;
            }
            return true;
        });
        files = files.map(function (item) {
            var ext = pathModule.extname(item);
            var base = pathModule.basename(item, ext);
            return pathModule.join(path, base);
        });
        var toLoads = [];
        for (var _i = 0, files_1 = files; _i < files_1.length; _i++) {
            var file = files_1[_i];
            if (toLoads.indexOf(file) >= 0)
                continue;
            toLoads.push(file);
        }
        for (var _a = 0, toLoads_1 = toLoads; _a < toLoads_1.length; _a++) {
            var file = toLoads_1[_a];
            this.loadService(file);
        }
    };
    ServiceContext.prototype.loadService = function (path) {
        var result = require(path) || {};
        for (var name_2 in result) {
            var obj = result[name_2];
            if (this.isServiceConstructor(obj)) {
                var Cons = obj;
                var s = new Cons(this.settings);
                if (!s.name) {
                    console.log(useColors.cyan("Detect service without name. Use constructor name:" + Cons.name));
                    s.name = Cons.name;
                }
                console.log("Load service " + useColors.green(s.name));
                this.register(s);
            }
        }
    };
    ServiceContext.prototype.isServiceConstructor = function (obj) {
        return typeof obj === "function" && obj.prototype["thisObjectIsRootServiceConstructor"];
    };
    ServiceContext.prototype.setup = function (finish) {
        var _this = this;
        if (finish === void 0) { finish = function () { return null; }; }
        this.finishSetup.push(finish);
        if (this.isSetup) {
            return;
        }
        this.isSetup = true;
        var initialSequence = this.dependencyManager.resolve();
        var timer = null;
        async.eachSeries(initialSequence, function (name, done) {
            var service = _this.services[name];
            if (!service) {
                done(new Error("Fail to find service name " + name));
                return;
            }
            if (service.isInitialized) {
                done();
                return;
            }
            service.services = _this.services;
            console.log("Initializing " + useColors.green(name));
            if (service.initialize) {
                if (service.initialize.length == 1) {
                    timer = setTimeout(function () {
                        console.error("Service " + name + " initializing timeout, exit program");
                        process.exit(1);
                    }, _this.maxInitializingTimeout);
                    service.initialize(function (err) {
                        clearTimeout(timer);
                        if (err) {
                            console.error(useColors.red("Fail to initialize service " + name + "."));
                            done(err);
                            return;
                        }
                        service.isInitialized = true;
                        _this.logInitialized(name);
                        done();
                    });
                }
                else {
                    service.initialize();
                    service.isInitialized = true;
                    _this.logInitialized(name);
                    done();
                }
            }
            else {
                // this service don't need initialization
                service.isInitialized = true;
                _this.logInitialized(name);
                done();
            }
        }, function (err) {
            _this.isSetup = false;
            if (err) {
                _this.finishSetup(err);
                return;
            }
            console.log(useColors.bgGreen("All service initialized"));
            _this.finishSetup();
        });
    };
    ServiceContext.prototype.logInitialized = function (name) {
        //console.log(`Service ${useColors.green(name)} Initialized`)
    };
    return ServiceContext;
}());
exports.ServiceContext = ServiceContext;
var Service = /** @class */ (function () {
    function Service(_defaultSettings, settings) {
        this._defaultSettings = _defaultSettings;
        this.dependencies = [];
        this.befores = [];
        this.isInitialized = false;
        this.settings = {};
        this.buildOption(_defaultSettings, settings);
    }
    Service.prototype.initialize = function (done) {
        done();
    };
    Service.prototype.buildOption = function (df, config) {
        this.settings = Leaf.Util.buildOption(df, config);
    };
    return Service;
}());
exports.Service = Service;
Service.prototype["thisObjectIsRootServiceConstructor"] = true;
//# sourceMappingURL=service.js.map